from flask import render_template, redirect, request, url_for, flash, session, abort, current_app
from flask_login import login_required, login_user, logout_user, current_user
from markupsafe import Markup, escape
from urllib.parse import urlencode
import secrets
import requests

from . import auth
from .forms import LoginForm, TwoFAForm, RegistrationForm, ChangePasswordForm, PasswordResetRequestForm, PasswordResetForm

from decorators import admin_required, permission_required
from mailsender import send_email
from models import db, User

from io import BytesIO
import pyqrcode
import sys

@auth.before_app_request
def before_request():
    #print('session: %s' % str(session))
    if current_user.is_authenticated:
        current_user.ping()
        #print('request for {} from {}#{}'.format(request.endpoint, current_user.email, current_user.id))
        if not current_user.confirmed and request.endpoint[:5] != 'auth.' and request.endpoint != 'static':
            print(request.endpoint)
            return redirect(url_for('auth.unconfirmed'))

@auth.route('/unconfirmed')
def unconfirmed():
    if current_user.is_anonymous or current_user.confirmed:
        return redirect(url_for('main.index'))
    return render_template('auth/unconfirmed.html')

@auth.route('/register', methods=['GET', 'POST'])
def register():
    page = { 'title': 'Register' }
    form = RegistrationForm()
    if form.validate_on_submit():
        user = User(email=form.email.data, password=form.password.data)
        db.session.add(user)
        db.session.commit()
        token = user.generate_confirmation_token()
        send_email(user.email, 'ForestNet Registration', 'auth/email/confirm', user=user, token=token)
        #notify admin
        newip = request.remote_addr
        if request.headers.getlist("X-Forwarded-For"):
            newip = request.headers.getlist("X-Forwarded-For")[0]
        else:
            newip = request.remote_addr
        send_email(current_app.config['MAIL_USERNAME'], user.email + ' registered!', 'auth/email/adm_regnotify', user=user, ipaddr=newip )
        flash('Thank you for the registration! We have send you an activation email')
        return redirect(url_for('auth.login'))
    return render_template('auth/register.html', page=page, form=form)

@auth.route('/confirm/<token>')
@login_required
def confirm(token):
    if current_user.confirmed:
        return redirect(url_for('main.index'))
    if current_user.confirm(token):
        flash('Your account is confirmed!')
    else:
        flash('Confirmation time expired!')
    return redirect(url_for('main.index'))

@auth.route('/confirm')
@login_required
def resend_confirmation():
    token = current_user.generate_confirmation_token()
    send_email(current_user.email, 'Confirm_your_account',
               'auth/email/confirm', user=current_user, token=token)
    flash('New confirmation code was sent.')
    return redirect(url_for('main.index'))

@auth.route('/authorize/<provider>')
def oauth2_authorize(provider):
    if not current_user.is_anonymous:
        return redirect(url_for('main.index'))

    provider_data = current_app.config['OAUTH2_PROVIDERS'].get(provider)
    if provider_data is None:
        abort(404)

    # generate a random string for the state parameter
    session['oauth2_state'] = secrets.token_urlsafe(16)

    # create a query string with all the OAuth2 parameters
    qs = urlencode({
        'client_id': provider_data['client_id'],
        'redirect_uri': url_for('auth.oauth2_callback', provider=provider,
                                _external=True, _scheme='https'),
        'response_type': 'code',
        'scope': ' '.join(provider_data['scopes']),
        'state': session['oauth2_state'],
    })

    # redirect the user to the OAuth2 provider authorization URL
    return redirect(provider_data['authorize_url'] + '?' + qs)

@auth.route('/callback/<provider>')
def oauth2_callback(provider):
    if not current_user.is_anonymous:
        return redirect(url_for('main.index'))

    provider_data = current_app.config['OAUTH2_PROVIDERS'].get(provider)
    if provider_data is None:
        abort(404)

    # if there was an authentication error, flash the error messages and exit
    if 'error' in request.args:
        for k, v in request.args.items():
            if k.startswith('error'):
                flash(f'{k}: {v}')
        return redirect(url_for('main.index'))

    # make sure that the state parameter matches the one we created in the
    # authorization request
    if request.args['state'] != session.get('oauth2_state'):
        print('oauth2 missing arg state!', file=sys.stderr)
        abort(401)

    # make sure that the authorization code is present
    if 'code' not in request.args:
        print('oauth missing code', file=sys.stderr)
        abort(401)

    # exchange the authorization code for an access token
    response = requests.post(provider_data['token_url'], data={
        'client_id': provider_data['client_id'],
        'client_secret': provider_data['client_secret'],
        'code': request.args['code'],
        'grant_type': 'authorization_code',
        'redirect_uri': url_for('auth.oauth2_callback', provider=provider,
                                _external=True, _scheme='https'),
    }, headers={'Accept': 'application/json'})
    if response.status_code != 200:
        print('oauth response code is not 200', file=sys.stderr)
        abort(401)
    oauth2_token = response.json().get('access_token')
    if not oauth2_token:
        print('oauth no access token', file=sys.stderr)
        abort(401)

    # use the access token to get the user's email address
    response = requests.get(provider_data['userinfo']['url'], headers={
        'Authorization': 'Bearer ' + oauth2_token,
        'Accept': 'application/json',
    })
    if response.status_code != 200:
        print('oauth no user info', file=sys.stderr)
        abort(401)
    email = provider_data['userinfo']['email'](response.json())

    # find or create the user in the database
    user = db.session.scalar(db.select(User).where(User.email == email))
    if user is None:
        #user = User(email=email, username=email.split('@')[0])
        user = User(email=email, confirmed=True)
        db.session.add(user)
        db.session.commit()

    # log the user in
    login_user(user)
    #return redirect(url_for('main.index'))
    flash('Last Login: {}'.format(user.last_seen.strftime("%a %d %B %Y %H:%M")))
    return redirect(request.args.get('next') or url_for('panel.dashboard'))

@auth.route('/login', methods=['GET', 'POST'])
def login():
    page = { 'title': 'Login' }
    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=form.email.data).first()

        if user is not None and user.verify_password(form.password.data):
            if user.active == False:
                flash('User disabled.')
                return redirect(url_for('main.index'))

            if user.twofactor:
                # redirect to the two-factor auth page, passing username in session
                session['email'] = user.email
                session['memberberry'] = form.remember_me.data
                return redirect(url_for('auth.twofactor'))

            login_user(user, form.remember_me.data)
            previp = user.last_ip
            if request.headers.getlist("X-Forwarded-For"):
                lastip = request.headers.getlist("X-Forwarded-For")[0]
            else:
                lastip = request.remote_addr
            user.last_ip = lastip
            db.session.add(user)
            db.session.commit()
            send_email(current_app.config['MAIL_USERNAME'], user.email + ' logged in.', 'auth/email/adm_loginnotify', user=user, ipaddr=lastip )
            flash('Last Login: {}'.format(user.last_seen.strftime("%a %d %B %Y %H:%M")))
            return redirect(request.args.get('next') or url_for('panel.dashboard'))
        else:
            flash('Invalid username or password.')

    return render_template('auth/login.html', page=page, form=form)

@auth.route('/twofactor', methods=['GET', 'POST'])
def twofactor():
    if 'email' not in session:
        abort(404)
    if 'memberberry' not in session:
        abort(404)
    page = { 'title': '2-Factor Login' }
    form = TwoFAForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=session['email']).first()
        del session['email']

        if user is not None and user.verify_totp(form.token.data):
            print('remember: ' + str(session['memberberry']))
            login_user(user, session['memberberry'])
            del session['memberberry']

            if request.headers.getlist("X-Forwarded-For"):
                lastip = request.headers.getlist("X-Forwarded-For")[0]
            else:
                lastip = request.remote_addr
            user.last_ip = lastip
            db.session.add(user)
            db.session.commit()
            #send_email(current_app.config['MAIL_USERNAME'], user.email + ' logged in.', 'auth/email/adm_loginnotify', user=user, ipaddr=lastip )
            return redirect(request.args.get('next') or url_for('panel.dashboard'))
        else:
            flash('Invalid token.')
    return render_template('auth/2fa.html', page=page, form=form)

@auth.route('/qrcode')
@login_required
def qrcode():
    #if 'email' not in session:
    #    abort(404)
    #user = User.query.filter_by(email=session['email']).first()
    #if user is None:
    #    abort(404)

    # for added security, remove username from session
    #del session['email']

    # render qrcode for FreeTOTP
    url = pyqrcode.create(current_user.get_totp_uri())
    stream = BytesIO()
    url.svg(stream, scale=6)
    svg_secret = Markup(stream.getvalue().decode('utf-8'))
    otp_secret = current_user.get_otp_secret()
    # since this page contains the sensitive qrcode, make sure the browser
    # does not cache it
    return render_template('auth/qrcode.html', svg=svg_secret, otp=otp_secret), 200, {
        'Cache-Control': 'no-cache, no-store, must-revalidate',
        'Pragma': 'no-cache',
        'Expires': '0'}

@auth.route('/change-password', methods=['GET', 'POST'])
@login_required
def change_password():
    form = ChangePasswordForm()
    if form.validate_on_submit():
        if current_user.verify_password(form.old_password.data):
            current_user.password = form.password.data
            db.session.add(current_user)
            db.session.commit()
            flash('Your password was changed')
            return redirect(url_for('main.index'))
        else:
            flash('Wrong password.')
    return render_template("auth/change_password.html", form=form)

@auth.route('/reset', methods=['GET', 'POST'])
def password_reset_request():
    if not current_user.is_anonymous:
        return redirect(url_for('main.index'))
    form = PasswordResetRequestForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=form.email.data).first()
        if user:
            token = user.generate_reset_token()
            send_email(user.email, 'Reset Your Password',
                       'auth/email/reset_password',
                       user=user, token=token,
                       next=request.args.get('next'))
            flash('An email with instructions to reset your password has been sent to you.')
        return redirect(url_for('auth.login'))
    return render_template('auth/reset_password.html', form=form)

@auth.route('/reset/<token>', methods=['GET', 'POST'])
def password_reset(token):
    if not current_user.is_anonymous:
        return redirect(url_for('main.index'))
    form = PasswordResetForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=form.email.data).first()
        if user is None:
            return redirect(url_for('main.index'))
        if user.reset_password(token, form.password.data):
            flash('Your password has been updated.')
            return redirect(url_for('auth.login'))
        else:
            return redirect(url_for('main.index'))
    return render_template('auth/reset_password.html', form=form)

@auth.route("/logout", methods=['GET'])
@login_required
def logout():
    logout_user()
    flash('You have logged out')
    return redirect(url_for('main.index'))

